/* ------------------------------- extract ------------------------------- */

/*
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <stdarg.h>

/* All Borland C/C++ versions */
#ifdef __TURBOC__
	 #define MSDOS 1
	 #include <io.h>
#else
	/* Microsoft */
	 #define MSDOS 1
#endif

#define BASE_HEADER_SIZE   19
#define CRC_MASK           0xFFFFFFFFL
#define CRC32_POLYNOMIAL   0xEDB88320L

/*
 * The only data structure used inside the CAR file is the header block.
 * Each file is preceded by a header, stored in a portable format.
 * The header is read into and out of the structure defined below.
 * The CAR file is structured as a series of header/data sequences, with
 * the EOF being denoted as a header with a file name length of 0.  Note
 * that the length of each header will vary depending on the length of
 * the file name.
 */

#ifndef FILENAME_MAX
#define FILENAME_MAX 128
#endif

typedef struct header {
	 char file_name[ FILENAME_MAX ];
	 char compression_method;
	 unsigned long original_size;
	 unsigned long compressed_size;
	 unsigned long original_crc;
	 unsigned long header_crc;
} HEADER;

/*
 * Fatal error exit
void FatalError(char *format, char *arg)
{
	printf(format,arg);
	exit(1);
}
 */
/*
 * Local function prototypes
 */

void BuildCRCTable( void );
unsigned long CalculateBlockCRC32( unsigned int count,
	unsigned long crc,
	void *buffer );
unsigned long UpdateCharacterCRC32( unsigned long crc, int c );
void SkipOverFileFromInputCar( void );
int ReadFileHeader( void );
unsigned long UnpackUnsignedData( int number_of_bytes, unsigned char *buffer );
int extract( FILE *in, FILE *out, char *name);
unsigned long Unstore( FILE *destination );
unsigned long LZSSExpand( FILE *destination );
void InitTree( int r );
void ContractNode( int old_node, int new_node );
void ReplaceNode( int old_node, int new_node );
int FindNextNode( int node );
void DeleteString( int p );
int AddString( int new_node, int *match_position );
void InitInputBuffer( void );
int InputBit( void );

static FILE *InputCarFile;           /* The input CAR file. */

static HEADER Header;             /* The Header block for the file  */
						/* presently being operated on    */

static unsigned long Ccitt32Table[ 256 ];    /* This array holds the CRC */
						/* table used to calculate the 32 */
						/* bit CRC values.                */
/*
 * This routine simply builds the coefficient table used to calculate
 * 32 bit CRC values throughout this program.  The 256 long word table
 * has to be set up once when the program starts.  Alternatively, the
 * values could be hard coded in, which would offer a miniscule improvement
 * in overall performance of the program.
 */

void BuildCRCTable()
{
    int i;
    int j;
    unsigned long value;

    for ( i = 0; i <= 255 ; i++ ) {
	value = i;
	for ( j = 8 ; j > 0; j-- ) {
	    if ( value & 1 )
		value = ( value >> 1 ) ^ CRC32_POLYNOMIAL;
	    else
		value >>= 1;
	}
	Ccitt32Table[ i ] = value;
    }
}

/*
 * This is the routine used to calculate the 32 bit CRC of a block of data.
 * This is done by processing the input buffer using the coefficient table
 * that was created when the program was initialized.  This routine takes
 * an input value as a seed, so that a running calculation of the CRC can
 * be used as blocks are read and written.
 */

unsigned long CalculateBlockCRC32( count, crc, buffer )
unsigned int count;
unsigned long crc;
void *buffer;
{
    unsigned char *p = (unsigned char *) buffer;
    unsigned long temp1;
    unsigned long temp2;

    while ( count-- != 0 ) {
	temp1 = ( crc >> 8 ) & 0x00FFFFFFL;
	temp2 = Ccitt32Table[ ( (int) crc ^ *p++ ) & 0xff ];
	crc = temp1 ^ temp2;
    }
    return( crc );
}

/*
 * If I/O is being done on a byte by byte basis, as is the case with the
 * LZSS code, it is easier to calculate the CRC of a byte at a time instead
 * of a block at a time.  This routine performs that function, once again
 * taking a crc value as input, so that this can be used to perform on the
 * fly calculations.  In situations where performance is critical, this
 * routine could easily be recoded as a macro.
 */

unsigned long UpdateCharacterCRC32( crc, c )
unsigned long crc;
int c;
{
    unsigned long temp1;
    unsigned long temp2;

    temp1 = ( crc >> 8 ) & 0x00FFFFFFL;
    temp2 = Ccitt32Table[ ( (int) crc ^ c ) & 0xff ];
    crc = temp1 ^ temp2;
    return( crc );
}

/*
 * This routine is where all the information about the next file in
 * the archive is read in.  The data is read into the global Header
 * structure.  To preserve portability of CAR files across systems,
 * the data in each file header is packed into an unsigned char array
 * before it is written out to the file.  To read this data back in
 * to the Header structure, we first read it into another unsigned
 * character array, then employ an unpacking routine to convert that
 * data into ints and longs.  This helps us avoid problems with
 * big/little endian conflicts, as well as incompatibilities in structure
 * packing, which show up even between different compilers targetted for
 * the same architecture.
 *
 * To avoid causing any additional confusion, the data members for the
 * header structure are at least stored in exactly the same order as
 * they appear in the structure definition.  The primary difference is
 * that the entire file name character array is not stored, which would
 * waste a lot of space.  Instead, we just store the number of characters
 * in the name, including the null termination character.  The file name
 * serves the additional purpose of identifying the end of the CAR file
 * with a file name length of 0 bytes.
 */

int ReadFileHeader()
{
/* Returns: 0 = success, -1 = EOF, else error */
   unsigned char header_data[ 17 ];
   unsigned long header_crc;
   int i;
   int c;
   for ( i = 0 ; ; ) {
      c = getc(InputCarFile);
      Header.file_name[ i ] = (char) c;
      if ( c == '\0' ) break;
      if ( ++i == FILENAME_MAX )
/*         FatalError( "File name exceeded maximum in header",""); */
           return(2);
   }
   if ( i == 0 ) return(-1);
   header_crc = CalculateBlockCRC32( i + 1, CRC_MASK, Header.file_name );
   fread( header_data, 1, 17, InputCarFile );
   Header.compression_method = (char) UnpackUnsignedData(1,header_data+0);
   Header.original_size      = UnpackUnsignedData( 4, header_data + 1  );
   Header.compressed_size    = UnpackUnsignedData( 4, header_data + 5  );
   Header.original_crc       = UnpackUnsignedData( 4, header_data + 9  );
   Header.header_crc         = UnpackUnsignedData( 4, header_data + 13 );
   header_crc = CalculateBlockCRC32( 13, header_crc, header_data );
   header_crc ^= CRC_MASK;
   if ( Header.header_crc != header_crc )
/*    FatalError( "Header checksum error for file %s", Header.file_name ); */
      return(2);
   return(0);  /* normal return */
}
/*
 * When the main processing loop reads in a header, it checks to see
 * if it is going to copy that file either to the OutputCarFile or expand
 * it.  If neither is going to happen, we need to skip past this file and
 * go on to the next header.  This can be done by seeking past the
 * compressed file.  Since the compressed size is stored in the header
 * information, it is easy to do.  Note that this routine assumes that the
 * file pointer has not been modified since the header was read in.  This
 * means it should be located at the first byte of the compressed data.
 */

void SkipOverFileFromInputCar()
{
	fseek( InputCarFile, Header.compressed_size, SEEK_CUR );
}

/*
 * This routine is used to transform packed characters into unsigned
 * integers.  Its only purpose is to convert packed character data
 * into integers and longs.
 */

unsigned long UnpackUnsignedData( number_of_bytes, buffer )
int number_of_bytes;
unsigned char *buffer;
{
	 unsigned long result;
	 int shift_count;

	 result = 0;
	 shift_count = 0;
	 while ( number_of_bytes-- > 0 ) {
	result |= (unsigned long) *buffer++ << shift_count;
		  shift_count += 8;
	 }
	 return( result );
}

/*
 * Extract is called for extracting the named file from the already
 * opened CAR file and write it to the alread opened output file.
 * Returns 0 on success, else 1.
 */

int extract(FILE *CARfile, FILE *out, char *name)
{
   unsigned long crc;

   /* Build CRC table for extraction operation */
   BuildCRCTable();
   /* Find the named file in the CAR file */
   InputCarFile = CARfile;

   /* Read CAR file until name is found, or EOF and not found */
   while (ReadFileHeader() == 0) {
      if(strcmp(name, Header.file_name) == 0) {
/*         printf("%-20s ", Header.file_name ); */
         switch (Header.compression_method) {
            case 1 :
               crc = Unstore(out);
               break;
            case 2 :
               crc = LZSSExpand(out);
               break;
            default :
/*             printf("Unknown method: %c\n",Header.compression_method );  */
               return(1);
         }
         if (crc != Header.original_crc) {
/*          printf("CRC error reading data\n"); */
            return(3);
         }
         return(0);
      }
      SkipOverFileFromInputCar();
   }
/* printf("Source file %s not found\n",name); */
   return(1);
}
/*
 * Extract uncompressed file from CAR file.
 */
unsigned long Unstore(FILE *destination)
{
   unsigned long crc;
   unsigned int count;
   unsigned char buffer[ 256 ];
   int pacifier;
   pacifier = 0;
   crc = CRC_MASK;
   while ( Header.original_size != 0 ) {
      if ( Header.original_size > 256 )
         count = 256;
   else
      count = (int) Header.original_size;
      if(fread(buffer, 1, count, InputCarFile) != count ) {
/*        FatalError("Can't read from input CAR file",""); */
          return(~Header.original_crc);
      }
      if(fwrite(buffer, 1, count, destination) != count ) {
/*        printf("Error writing to output file"); */
          return(~Header.original_crc);
      }
      crc = CalculateBlockCRC32( count, crc, buffer );
      if(destination != stdout && ( pacifier++ & 15 ) == 0 ) putchar( '.');
      Header.original_size -= count;
   }
   putchar('\n');
   return( crc ^ CRC_MASK );
}

/*
 * The second set of compression routines are found here.  These
 * routines implement LZSS compression and expansion using 12 bit
 * index pointers and 4 bit match lengths.  These values were
 * specifically chosen because they allow for "blocked I/O".  Because
 * of their values, we can pack match/length pairs into pairs of
 * bytes, with characters that don't have matches going into single
 * bytes.  This helps increase I/O since single bit input and
 * output does not have to be employed.  Other than this single change,
 * this code is identical to the LZSS code used earlier in the book.
 */

/*
 * Various constants used to define the compression parameters.  The
 * INDEX_BIT_COUNT tells how many bits we allocate to indices into the
 * text window.  This directly determines the WINDOW_SIZE.  The
 * LENGTH_BIT_COUNT tells how many bits we allocate for the length of
 * an encode phrase. This determines the size of the look ahead buffer.
 * The TREE_ROOT is a special node in the tree that always points to
 * the root node of the binary phrase tree.  END_OF_STREAM is a special
 * index used to flag the fact that the file has been completely
 * encoded, and there is no more data.  UNUSED is the null index for
 * the tree. MOD_WINDOW() is a macro used to perform arithmetic on tree
 * indices.
 *
 */

#define INDEX_BIT_COUNT      12
#define LENGTH_BIT_COUNT     4
#define WINDOW_SIZE          ( 1 << INDEX_BIT_COUNT )
#define RAW_LOOK_AHEAD_SIZE  ( 1 << LENGTH_BIT_COUNT )
#define BREAK_EVEN           ( ( 1 + INDEX_BIT_COUNT + LENGTH_BIT_COUNT ) / 9 )
#define LOOK_AHEAD_SIZE      ( RAW_LOOK_AHEAD_SIZE + BREAK_EVEN )
#define TREE_ROOT            WINDOW_SIZE
#define END_OF_STREAM        0
#define UNUSED               0
#define MOD_WINDOW( a )      ( ( a ) & ( WINDOW_SIZE - 1 ) )

/*
 * These are the two global data structures used in this program.
 * The window[] array is exactly that, the window of previously seen
 * text, as well as the current look ahead text.  The tree[] structure
 * contains the binary tree of all of the strings in the window sorted
 * in order.
*/

unsigned char window[ WINDOW_SIZE ];

struct {
    int parent;
    int smaller_child;
    int larger_child;
} tree[ WINDOW_SIZE + 1 ];

/*
 * Since the tree is static data, it comes up with every node
 * initialized to 0, which is good, since 0 is the UNUSED code.
 * However, to make the tree really usable, a single phrase has to be
 * added to the tree so it has a root node.  That is done right here.
*/
void InitTree( r )
int r;
{
    int i;

    for ( i = 0 ; i < ( WINDOW_SIZE + 1 ) ; i++ ) {
        tree[ i ].parent = UNUSED;
        tree[ i ].larger_child = UNUSED;
        tree[ i ].smaller_child = UNUSED;
    }
    tree[ TREE_ROOT ].larger_child = r;
    tree[ r ].parent = TREE_ROOT;
    tree[ r ].larger_child = UNUSED;
    tree[ r ].smaller_child = UNUSED;
}

/*
 * This routine is used when a node is being deleted.  The link to
 * its descendant is broken by pulling the descendant in to overlay
 * the existing link.
 */
void ContractNode( old_node, new_node )
int old_node;
int new_node;
{
    tree[ new_node ].parent = tree[ old_node ].parent;
    if ( tree[ tree[ old_node ].parent ].larger_child == old_node )
        tree[ tree[ old_node ].parent ].larger_child = new_node;
    else
        tree[ tree[ old_node ].parent ].smaller_child = new_node;
    tree[ old_node ].parent = UNUSED;
}

/*
 * This routine is also used when a node is being deleted.  However,
 * in this case, it is being replaced by a node that was not previously
 * in the tree.
 */
void ReplaceNode( old_node, new_node )
int old_node;
int new_node;
{
    int parent;

    parent = tree[ old_node ].parent;
    if ( tree[ parent ].smaller_child == old_node )
        tree[ parent ].smaller_child = new_node;
    else
        tree[ parent ].larger_child = new_node;
    tree[ new_node ] = tree[ old_node ];
    tree[ tree[ new_node ].smaller_child ].parent = new_node;
    tree[ tree[ new_node ].larger_child ].parent = new_node;
    tree[ old_node ].parent = UNUSED;
}

/*
 * This routine is used to find the next smallest node after the node
 * argument.  It assumes that the node has a smaller child.  We find
 * the next smallest child by going to the smaller_child node, then
 * going to the end of the larger_child descendant chain.
*/
int FindNextNode( node )
int node;
{
    int next;

    next = tree[ node ].smaller_child;
    while ( tree[ next ].larger_child != UNUSED )
        next = tree[ next ].larger_child;
    return( next );
}

/*
 * This routine performs the classic binary tree deletion algorithm.
 * If the node to be deleted has a null link in either direction, we
 * just pull the non-null link up one to replace the existing link.
 * If both links exist, we instead delete the next link in order, which
 * is guaranteed to have a null link, then replace the node to be deleted
 * with the next link.
 */
void DeleteString( p )
int p;
{
    int  replacement;

    if ( tree[ p ].parent == UNUSED )
        return;
    if ( tree[ p ].larger_child == UNUSED )
        ContractNode( p, tree[ p ].smaller_child );
    else if ( tree[ p ].smaller_child == UNUSED )
        ContractNode( p, tree[ p ].larger_child );
    else {
        replacement = FindNextNode( p );
        DeleteString( replacement );
        ReplaceNode( p, replacement );
    }
}

/*
 * This where most of the work done by the encoder takes place.  This
 * routine is responsible for adding the new node to the binary tree.
 * It also has to find the best match among all the existing nodes in
 * the tree, and return that to the calling routine.  To make matters
 * even more complicated, if the new_node has a duplicate in the tree,
 * the old_node is deleted, for reasons of efficiency.
 */

int AddString( new_node, match_position )
int new_node;
int *match_position;
{
    int i;
    int test_node;
    int delta;
    int match_length;
    int *child;

    if ( new_node == END_OF_STREAM )
        return( 0 );
    test_node = tree[ TREE_ROOT ].larger_child;
    match_length = 0;
    for ( ; ; ) {
        for ( i = 0 ; i < LOOK_AHEAD_SIZE ; i++ ) {
            delta = window[ MOD_WINDOW( new_node + i ) ] -
                    window[ MOD_WINDOW( test_node + i ) ];
            if ( delta != 0 )
                break;
        }
        if ( i >= match_length ) {
            match_length = i;
            *match_position = test_node;
            if ( match_length >= LOOK_AHEAD_SIZE ) {
                ReplaceNode( test_node, new_node );
                return( match_length );
            }
        }
        if ( delta >= 0 )
            child = &tree[ test_node ].larger_child;
        else
            child = &tree[ test_node ].smaller_child;
        if ( *child == UNUSED ) {
            *child = new_node;
            tree[ new_node ].parent = test_node;
            tree[ new_node ].larger_child = UNUSED;
            tree[ new_node ].smaller_child = UNUSED;
            return( match_length );
        }
        test_node = *child;
    }
}

/*
 * This section of code and data makes up the blocked I/O portion of
 * the program.  Every token output consists of a single flag bit, followed
 * by either a single character or a index/length pair.  The flag bits
 * are stored in the first byte of a buffer array, and the characters
 * and index/length pairs are stored sequentially in the remaining
 * positions in the array.  After every eight output operations, the
 * first character of the array is full of flag bits, so the remaining
 * bytes stored in the array can be output.  This can be done with a
 * single fwrite() operation, making for greater efficiency.
 *
 * All that is needed to implement this is a few routines, plus three
 * data objects, which follow below.  The buffer has the flag bits
 * packed into its first character, with the remainder consisting of
 * the characters and index/length pairs, appearing in the order they
 * were output.  The FlagBitMask  is used to indicate where the next
 * flag bit will go when packed into DataBuffer[ 0 ].  Finally, the
 * BufferOffset is used to indicate where the next token will be stored
 * in the buffer.
 */

char DataBuffer[ 17 ];
int FlagBitMask;
unsigned int BufferOffset;

/*
 * To initialize the output buffer, we set the FlagBitMask to the first
 * bit position, can clear DataBuffer[0], which will hold all the
 * Flag bits.  Finally, the BufferOffset is set to 1, which is where the
 * first character or index/length pair will go.
 */

void InitOutputBuffer()
{
    DataBuffer[ 0 ] = 0;
    FlagBitMask = 1;
    BufferOffset = 1;
}

/*
 * The input process uses the same data structures as the blocked output
 * routines, but it is somewhat simpler, in that it doesn't actually have
 * to read in a whole block of data at once.  Instead, it just reads in
 * a single character full of flag bits into DataBuffer[0], and passes
 * individual bits back to the Expansion program when asked for them.
 * The expansion program is left to its own devices for reading in the
 * characters, indices, and match lengths.  They can be read in sequentially
 * using normal file I/O.
 */

void InitInputBuffer()
{
    FlagBitMask = 1;
    DataBuffer[ 0 ] = (char) getc( InputCarFile );
}

/*
 * When the Expansion program wants a flag bit, it calls this routine.
 * This routine has to keep track of whether or not it has run out of
 * flag bits.  If it has, it has to go back and reinitialize so as to
 * have a fresh set.
 */

int InputBit()
{
    if ( FlagBitMask == 0x100 )
        InitInputBuffer();
    FlagBitMask <<= 1;
    return( DataBuffer[ 0 ] & ( FlagBitMask >> 1 ) );
}

/*
 * This is the expansion routine for the LZSS algorithm.  All it has
 * to do is read in flag bits, decide whether to read in a character or
 * a index/length pair, and take the appropriate action.  It is responsible
 * for keeping track of the crc of the output data, and must return it
 * to the calling routine, for verification.
 */

unsigned long LZSSExpand( output )
FILE *output;
{
    int i;
    int current_position;
    int c;
    int match_length;
    int match_position;
    unsigned long crc;
    unsigned long output_count;

    output_count = 0;
    crc = CRC_MASK;
    InitInputBuffer();
    current_position = 1;
    while ( output_count < Header.original_size ) {
        if ( InputBit() ) {
            c = getc( InputCarFile );
            putc( c, output );
            output_count++;
            crc = UpdateCharacterCRC32( crc, c );
            window[ current_position ] = (unsigned char) c;
            current_position = MOD_WINDOW( current_position + 1 );
            if ( current_position == 0 && output != stdout )
                putchar('.');
        } else {
            match_length = getc( InputCarFile );
            match_position = getc( InputCarFile );
            match_position |= ( match_length & 0xf ) << 8;
            match_length >>= 4;
            match_length += BREAK_EVEN;
            output_count += match_length + 1;
            for ( i = 0 ; i <= match_length ; i++ ) {
                c = window[ MOD_WINDOW( match_position + i ) ];
                putc( c, output );
                crc = UpdateCharacterCRC32( crc, c );
                window[ current_position ] = (unsigned char) c;
                current_position = MOD_WINDOW( current_position + 1 );
                if ( current_position == 0 && output != stdout )
                    putc( '.', stderr );
            }
        }
    }
    putchar('\n');
    return( crc ^ CRC_MASK );
}
/************************** End of CARMAN.C *************************/
